==> fred-inquisitor-1.0.0rc1.tar.gz: fred-inquisitor-1.0.0rc1/inquisitor/oidc.py <==
"""OIDC verification endpoints."""
import json
from urllib.parse import urlencode

from fastapi import APIRouter, Depends, HTTPException, Request, Security, status
from fastapi.responses import RedirectResponse
from fred_types import BaseModel, ContactId
from grill import Logger
from oic.oic import Client as OIDCClient
from oic.oic.message import AuthorizationResponse, Claims, ClaimsRequest
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from regal import Address
from verifier import ContactData, VerificationClient, VerificationId, VerificationResult
from verifier.exceptions import (
    IncorrectStatus,
    InvalidData,
    MethodNotApplicable,
    VerificationDoesNotExist,
)

from .auth import get_contact_id, get_verification_id
from .constants import (
    COMMON_RESPONSES,
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    Scope,
    VerificationMethod,
)
from .models import Error
from .session import OIDCSession, Session, get_session
from .settings import Settings, get_logger, get_settings, get_verification_client
from .utils import get_source_ip

router = APIRouter(prefix="/oidc")

USERINFO_CLAIMS = {
    "name",
    "mojeid_address_def_street",
    "mojeid_address_def_street2",
    "mojeid_address_def_street3",
    "mojeid_address_def_city",
    "mojeid_address_def_state",
    "mojeid_address_def_postal_code",
    "mojeid_address_def_country",
    "mojeid_nia",
    "mojeid_valid",
    "mojeid_ident_vat",
    "mojeid_organization",
}


def get_oidc_client(request: Request, settings: Settings = Depends(get_settings)) -> OIDCClient:
    """Get OIDC client."""
    client = OIDCClient(client_authn_method=CLIENT_AUTHN_METHOD)
    client.provider_config(settings.oidc.issuer)
    client.client_id = settings.oidc.client_id
    client.client_secret = settings.oidc.client_secret
    client.redirect_uris = [str(request.url_for("oidc_token"))]
    return client


@router.get("/auth/")
async def oidc_auth(
    redirect: str = "/",
    source_ip: str | None = Depends(get_source_ip),
    session: Session = Security(get_session, scopes=[Scope.WRITE]),
    client: OIDCClient = Depends(get_oidc_client),
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_id: ContactId = Depends(get_contact_id),
    logger: Logger = Depends(get_logger),
) -> RedirectResponse:
    """Initiate OIDC authentication."""
    log_entry = logger.create(
        LogEntryType.USE_METHOD,
        source_ip=source_ip,
        properties={"method": VerificationMethod.MOJEID},
        references={
            LogReferenceType.VERIFICATION: verification_id,
            LogReferenceType.CONTACT: contact_id,
        },
    )
    with log_entry:
        session.oidc = OIDCSession(redirect_url=redirect)
        args = {
            "client_id": client.client_id,
            "response_type": "code",
            "nonce": session.oidc.nonce,
            "redirect_uri": client.redirect_uris[0],
            "state": session.oidc.state,
            "claims": ClaimsRequest(
                userinfo=Claims(**{claim: {"essential": True} for claim in USERINFO_CLAIMS})
            ),
        }

        auth_req = client.construct_AuthorizationRequest(request_args=args)
        login_url = auth_req.request(client.authorization_endpoint)

        try:
            await verification_client.use_method(verification_id, method=VerificationMethod.MOJEID)
        except VerificationDoesNotExist:
            session.oidc.error = Error(detail="Verification not found")
            log_entry.properties["error"] = "Verification not found"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)
        except MethodNotApplicable:
            session.oidc.error = Error(detail="Method not applicable")
            log_entry.properties["error"] = "Method not applicable"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)
        except IncorrectStatus:
            session.oidc.error = Error(detail="Incorrect verification status")
            log_entry.properties["error"] = "Incorrect verification status"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)

        log_entry.result = LoggerResult.SUCCESS
        return RedirectResponse(login_url)


def userinfo_to_contact_data(userinfo: dict) -> ContactData:
    """Convert userinfo to contact data model."""
    ADDRESS_FIELDS = {
        "mojeid_address_def_street",
        "mojeid_address_def_street2",
        "mojeid_address_def_street3",
        "mojeid_address_def_city",
        "mojeid_address_def_postal_code",
        "mojeid_address_def_country",
    }
    if ADDRESS_FIELDS & set(userinfo.keys()):
        streets = [
            userinfo.get("mojeid_address_def_street"),
            userinfo.get("mojeid_address_def_street2"),
            userinfo.get("mojeid_address_def_street3"),
        ]
        address = Address(
            street=[s for s in streets if s is not None],
            city=userinfo.get("mojeid_address_def_city"),
            state_or_province=userinfo.get("mojeid_address_def_state", ""),
            postal_code=userinfo.get("mojeid_address_def_postal_code"),
            country_code=userinfo.get("mojeid_address_def_country"),
        )
    else:
        address = None

    organization = userinfo.get("mojeid_organization", "")

    return ContactData(
        name=userinfo.get("name", ""),
        organization=organization,
        company_registration_number=userinfo.get("mojeid_ident_vat", "" if organization else None),
        address=address,
    )


@router.get("/token/")
async def oidc_token(
    request: Request,
    source_ip: str | None = Depends(get_source_ip),
    session: Session = Security(get_session, scopes=[Scope.WRITE]),
    settings: Settings = Depends(get_settings),
    client: OIDCClient = Depends(get_oidc_client),
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_id: ContactId = Depends(get_contact_id),
    logger: Logger = Depends(get_logger),
) -> RedirectResponse:
    """Finalize OIDC authentication.

    Attempt to verify contact data if the contact is trustworthy.
    """
    response = urlencode(request.query_params)
    aresp = client.parse_response(AuthorizationResponse, info=response, sformat="urlencoded")

    log_entry = logger.create(
        LogEntryType.VERIFY_CONTACT_DATA,
        source_ip=source_ip,
        properties={"method": VerificationMethod.MOJEID},
        content=json.dumps(dict(aresp), indent=4, ensure_ascii=False),
        references={
            LogReferenceType.VERIFICATION: verification_id,
            LogReferenceType.CONTACT: contact_id,
        },
    )
    with log_entry:
        if aresp["state"] != session.oidc.state:
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(status.HTTP_401_UNAUTHORIZED)

        # Get access token
        args = {"code": aresp["code"]}
        client.do_access_token_request(
            state=session.oidc.state,
            request_args=args,
            authn_method="client_secret_post",
        )

        # Get user info
        userinfo = client.do_user_info_request(state=session.oidc.state)
        session.oidc.contact_data = userinfo_to_contact_data(userinfo)
        session.oidc.mojeid_nia = userinfo.get("mojeid_nia", False)
        session.oidc.mojeid_validation = userinfo.get("mojeid_valid", False)

        log_entry.content = json.dumps(dict(userinfo), indent=4, ensure_ascii=False)

        if session.oidc.is_trustworthy:
            try:
                _, result = await verification_client.verify_contact_data(
                    verification_id,
                    method=VerificationMethod.MOJEID,
                    data=session.oidc.contact_data,
                )
                session.oidc.verification_result = result
            except VerificationDoesNotExist:
                session.oidc.error = Error(detail="Verification not found")
                log_entry.properties["error"] = "Verification not found"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except MethodNotApplicable:
                session.oidc.error = Error(detail="Method not applicable")
                log_entry.properties["error"] = "Method not applicable"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except IncorrectStatus:
                session.oidc.error = Error(detail="Incorrect verification status")
                log_entry.properties["error"] = "Incorrect verification status"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except InvalidData:
                session.oidc.error = Error(detail="Invalid data")
                log_entry.properties["error"] = "Invalid data"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            else:
                log_entry.result = LoggerResult.SUCCESS
        else:
            log_entry.result = LoggerResult.FAIL

        return RedirectResponse(session.oidc.redirect_url)


class OIDCAuthResult(BaseModel):
    """OIDC auth result."""

    contact_data: ContactData
    mojeid_nia: bool
    mojeid_validation: bool
    is_legal_person: bool
    verification_result: VerificationResult | None = None


@router.get(
    "/last/",
    responses={**COMMON_RESPONSES, 400: {"description": "Operation error", "model": Error}},
)
async def get_last_oidc_result(
    session: Session = Security(get_session, scopes=[Scope.READ])
) -> OIDCAuthResult:
    """Return last OIDC authentication result."""
    if session.oidc.error:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail=session.oidc.error.detail)
    if not session.oidc.contact_data:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="No contact data available")
    return OIDCAuthResult(
        contact_data=session.oidc.contact_data,
        mojeid_nia=session.oidc.mojeid_nia,
        mojeid_validation=session.oidc.mojeid_validation,
        is_legal_person=session.oidc.is_legal_person,
        verification_result=session.oidc.verification_result,
    )

==> fred-inquisitor-1.2.0.tar.gz: fred-inquisitor-1.2.0/inquisitor/oidc.py <==
"""OIDC verification endpoints."""
import json
from urllib.parse import urlencode

from fastapi import APIRouter, Depends, HTTPException, Request, Security, status
from fastapi.responses import RedirectResponse
from fred_types import BaseModel, ContactId
from grill import Logger
from oic.oic import Client as OIDCClient
from oic.oic.message import AuthorizationResponse, Claims, ClaimsRequest
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from regal import Address
from verifier import ContactData, VerificationClient, VerificationId, VerificationResult
from verifier.exceptions import (
    IncorrectStatus,
    InvalidData,
    MethodNotApplicable,
    VerificationDoesNotExist,
)

from .constants import (
    COMMON_RESPONSES,
    LogEntryType,
    LoggerResult,
    LogReferenceType,
    Scope,
    VerificationMethod,
)
from .models import Error
from .session import OIDCSession, Session, get_session
from .settings import Settings, get_logger, get_settings, get_verification_client
from .token import get_contact_id, get_verification_id
from .utils import get_source_ip

router = APIRouter(prefix="/oidc")

USERINFO_CLAIMS = {
    "mojeid_username",
    "name",
    "mojeid_address_def_street",
    "mojeid_address_def_street2",
    "mojeid_address_def_street3",
    "mojeid_address_def_city",
    "mojeid_address_def_state",
    "mojeid_address_def_postal_code",
    "mojeid_address_def_country",
    "mojeid_nia",
    "mojeid_valid",
    "mojeid_ident_vat",
    "mojeid_organization",
}


def get_oidc_client(request: Request, settings: Settings = Depends(get_settings)) -> OIDCClient:
    """Get OIDC client."""
    client = OIDCClient(client_authn_method=CLIENT_AUTHN_METHOD)
    client.provider_config(settings.oidc.issuer)
    client.client_id = settings.oidc.client_id
    client.client_secret = settings.oidc.client_secret
    client.redirect_uris = [str(request.url_for("oidc_token"))]
    return client


@router.get("/auth/")
async def oidc_auth(
    redirect: str = "/",
    source_ip: str | None = Depends(get_source_ip),
    session: Session = Security(get_session, scopes=[Scope.WRITE]),
    client: OIDCClient = Depends(get_oidc_client),
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_id: ContactId = Depends(get_contact_id),
    logger: Logger = Depends(get_logger),
) -> RedirectResponse:
    """Initiate OIDC authentication."""
    log_entry = logger.create(
        LogEntryType.USE_METHOD,
        source_ip=source_ip,
        properties={"method": VerificationMethod.MOJEID},
        references={
            LogReferenceType.VERIFICATION: verification_id,
            LogReferenceType.CONTACT: contact_id,
        },
    )
    with log_entry:
        session.oidc = OIDCSession(redirect_url=redirect)
        args = {
            "client_id": client.client_id,
            "response_type": "code",
            "nonce": session.oidc.nonce,
            "redirect_uri": client.redirect_uris[0],
            "state": session.oidc.state,
            "claims": ClaimsRequest(
                userinfo=Claims(**{claim: {"essential": True} for claim in USERINFO_CLAIMS})
            ),
        }

        auth_req = client.construct_AuthorizationRequest(request_args=args)
        login_url = auth_req.request(client.authorization_endpoint)

        try:
            await verification_client.use_method(verification_id, method=VerificationMethod.MOJEID)
        except VerificationDoesNotExist:
            session.oidc.error = Error(detail="Verification not found")
            log_entry.properties["error"] = "Verification not found"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)
        except MethodNotApplicable:
            session.oidc.error = Error(detail="Method not applicable")
            log_entry.properties["error"] = "Method not applicable"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)
        except IncorrectStatus:
            session.oidc.error = Error(detail="Incorrect verification status")
            log_entry.properties["error"] = "Incorrect verification status"
            log_entry.result = LoggerResult.FAIL
            return RedirectResponse(redirect)

        log_entry.result = LoggerResult.SUCCESS
        return RedirectResponse(login_url)


def userinfo_to_contact_data(userinfo: dict) -> ContactData:
    """Convert userinfo to contact data model."""
    ADDRESS_FIELDS = {
        "mojeid_address_def_street",
        "mojeid_address_def_street2",
        "mojeid_address_def_street3",
        "mojeid_address_def_city",
        "mojeid_address_def_postal_code",
        "mojeid_address_def_country",
    }
    if ADDRESS_FIELDS & set(userinfo.keys()):
        streets = [
            userinfo.get("mojeid_address_def_street"),
            userinfo.get("mojeid_address_def_street2"),
            userinfo.get("mojeid_address_def_street3"),
        ]
        address = Address(
            street=[s for s in streets if s is not None],
            city=userinfo.get("mojeid_address_def_city"),
            state_or_province=userinfo.get("mojeid_address_def_state", ""),
            postal_code=userinfo.get("mojeid_address_def_postal_code"),
            country_code=userinfo.get("mojeid_address_def_country"),
        )
    else:
        address = None

    organization = userinfo.get("mojeid_organization", "")

    return ContactData(
        name=userinfo.get("name", ""),
        organization=organization,
        company_registration_number=userinfo.get("mojeid_ident_vat", "" if organization else None),
        address=address,
    )


@router.get("/token/")
async def oidc_token(
    request: Request,
    source_ip: str | None = Depends(get_source_ip),
    session: Session = Security(get_session, scopes=[Scope.WRITE]),
    settings: Settings = Depends(get_settings),
    client: OIDCClient = Depends(get_oidc_client),
    verification_id: VerificationId = Depends(get_verification_id),
    verification_client: VerificationClient = Depends(get_verification_client),
    contact_id: ContactId = Depends(get_contact_id),
    logger: Logger = Depends(get_logger),
) -> RedirectResponse:
    """Finalize OIDC authentication.

    Attempt to verify contact data if the contact is trustworthy.
    """
    response = urlencode(request.query_params)
    aresp = client.parse_response(AuthorizationResponse, info=response, sformat="urlencoded")

    log_entry = logger.create(
        LogEntryType.VERIFY_CONTACT_DATA,
        source_ip=source_ip,
        properties={"method": VerificationMethod.MOJEID},
        content=json.dumps(dict(aresp), indent=4, ensure_ascii=False),
        references={
            LogReferenceType.VERIFICATION: verification_id,
            LogReferenceType.CONTACT: contact_id,
        },
    )
    with log_entry:
        if aresp["state"] != session.oidc.state:
            log_entry.result = LoggerResult.FAIL
            raise HTTPException(status.HTTP_401_UNAUTHORIZED)

        if error := aresp.get("error"):
            # Some OIDC error occured
            log_entry.properties["error"] = error
            log_entry.result = LoggerResult.FAIL
            session.oidc.error = Error(detail=error)
            return RedirectResponse(session.oidc.redirect_url)

        # Get access token
        args = {"code": aresp["code"]}
        client.do_access_token_request(
            state=session.oidc.state,
            request_args=args,
            authn_method="client_secret_post",
        )

        # Get user info
        userinfo = client.do_user_info_request(state=session.oidc.state)
        session.oidc.contact_data = userinfo_to_contact_data(userinfo)
        session.oidc.mojeid_nia = userinfo.get("mojeid_nia", False)
        session.oidc.mojeid_validation = userinfo.get("mojeid_valid", False)
        session.oidc.mojeid_username = userinfo.get("mojeid_username", "")

        log_entry.content = json.dumps(dict(userinfo), indent=4, ensure_ascii=False)

        if session.oidc.is_trustworthy:
            try:
                _, result = await verification_client.verify_contact_data(
                    verification_id,
                    method=VerificationMethod.MOJEID,
                    data=session.oidc.contact_data,
                )
                session.oidc.verification_result = result
            except VerificationDoesNotExist:
                session.oidc.error = Error(detail="Verification not found")
                log_entry.properties["error"] = "Verification not found"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except MethodNotApplicable:
                session.oidc.error = Error(detail="Method not applicable")
                log_entry.properties["error"] = "Method not applicable"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except IncorrectStatus:
                session.oidc.error = Error(detail="Incorrect verification status")
                log_entry.properties["error"] = "Incorrect verification status"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            except InvalidData:
                session.oidc.error = Error(detail="Invalid data")
                log_entry.properties["error"] = "Invalid data"
                log_entry.result = LoggerResult.FAIL
                return RedirectResponse(session.oidc.redirect_url)
            else:
                log_entry.result = LoggerResult.SUCCESS
        else:
            log_entry.result = LoggerResult.FAIL

        return RedirectResponse(session.oidc.redirect_url)


class OIDCAuthResult(BaseModel):
    """OIDC auth result."""

    contact_data: ContactData
    mojeid_nia: bool
    mojeid_validation: bool
    mojeid_username: str
    is_legal_person: bool
    verification_result: VerificationResult | None = None


@router.get(
    "/last/",
    responses={**COMMON_RESPONSES, 400: {"description": "Operation error", "model": Error}},
)
async def get_last_oidc_result(
    session: Session = Security(get_session, scopes=[Scope.READ])
) -> OIDCAuthResult:
    """Return last OIDC authentication result."""
    if session.oidc.error:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail=session.oidc.error.detail)
    if not session.oidc.contact_data:
        raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="No contact data available")
    return OIDCAuthResult(
        contact_data=session.oidc.contact_data,
        mojeid_nia=session.oidc.mojeid_nia,
        mojeid_validation=session.oidc.mojeid_validation,
        mojeid_username=session.oidc.mojeid_username,
        is_legal_person=session.oidc.is_legal_person,
        verification_result=session.oidc.verification_result,
    )
